/*
 *  include/ub/ub_orphan.h
 *
 *  Copyright (C) 2005  SWsoft
 *  All rights reserved.
 *  
 *  Licensing governed by "linux/COPYING.SWsoft" file.
 *
 */

#ifndef __UB_ORPHAN_H_
#define __UB_ORPHAN_H_

#include <net/tcp.h>

#include "ub/beancounter.h"
#include "ub/ub_net.h"


static inline atomic_t *__ub_get_orphan_count_ptr(struct sock *sk)
{
#ifdef CONFIG_USER_RESOURCE
	if (sock_has_ubc(sk))
		return &sock_bc(sk)->ub->ub_orphan_count;
#endif
	return sk->sk_prot->orphan_count;
}

static inline void ub_inc_orphan_count(struct sock *sk)
{
	atomic_inc(__ub_get_orphan_count_ptr(sk));
}

static inline void ub_dec_orphan_count(struct sock *sk)
{
	atomic_dec(__ub_get_orphan_count_ptr(sk));
}

static inline int ub_get_orphan_count(struct sock *sk)
{
	return atomic_read(__ub_get_orphan_count_ptr(sk));
}

extern int __ub_too_many_orphans(struct sock *sk, int count);
static inline int ub_too_many_orphans(struct sock *sk, int count)
{
#ifdef CONFIG_USER_RESOURCE
	if (__ub_too_many_orphans(sk, count))
		return 1;
#endif
	return (ub_get_orphan_count(sk) > sysctl_tcp_max_orphans ||
		(sk->sk_wmem_queued > SOCK_MIN_SNDBUF &&
		 atomic_read(&tcp_memory_allocated) > sysctl_tcp_mem[2]));
}

#endif
